#![allow(clippy::manual_range_contains)]
#![warn(clippy::unreadable_literal, rust_2018_idioms)]
#![forbid(elided_lifetimes_in_paths, unsafe_code)]

use camino::Utf8PathBuf as PathBuf;
use clap::Parser;
use console::style;
use render_video::{
	cli::{ask, ask_time},
	preset::Preset,
	project::{Project, ProjectLecture, ProjectSource, Resolution},
	render::Renderer,
	time::parse_date
};
use std::fs;

#[derive(Debug, Parser)]
struct Args {
	/// The root directory of the project. It should contain the raw video file(s).
	#[clap(short = 'C', long, default_value = ".")]
	directory: PathBuf,

	/// The preset of the lecture. Can be a toml file or a known course slug.
	#[clap(short, long)]
	preset: String,

	#[cfg(feature = "mem_limit")]
	/// The memory limit for external tools like ffmpeg.
	#[clap(short, long, default_value = "12G")]
	mem_limit: String,

	/// Transcode the final video clip down to the minimum resolution specified. If not
	/// specified, the default value from the preset is used.
	#[clap(short, long)]
	transcode: Option<Resolution>,

	/// Transcode starts at this resolution, or the source resolution, whichever is lower.
	/// If not specified, the default value from the preset is used.
	#[clap(short = 'T', long)]
	transcode_start: Option<Resolution>,

	/// Treat the audio as stereo. By default, only one channel from the input stereo will
	/// be used, assuming either the other channel is backup or the same as the used.
	#[clap(short, long, default_value = "false")]
	stereo: bool
}

fn main() {
	let args = Args::parse();

	#[cfg(feature = "mem_limit")]
	{
		*(render_video::MEM_LIMIT.write().unwrap()) = args.mem_limit;
	}

	// process arguments
	let directory = args.directory.canonicalize_utf8().unwrap();
	let preset = Preset::find(&args.preset).unwrap();
	let course = preset.course;

	// let's see if we need to initialise the project
	let project_path = directory.join("project.toml");
	let mut project = if project_path.exists() {
		toml::from_slice(&fs::read(&project_path).unwrap()).unwrap()
	} else {
		let dirname = directory.file_name().unwrap();
		let date =
			parse_date(dirname).expect("Directory name is not in the expected format");

		let mut files = Vec::new();
		for entry in directory.read_dir_utf8().unwrap() {
			let entry = entry.unwrap();
			let name = entry.file_name();
			let lower = name.to_ascii_lowercase();
			if (lower.ends_with(".mp4")
				|| lower.ends_with(".mts")
				|| lower.ends_with(".mkv")
				|| lower.ends_with(".mov"))
				&& !entry.file_type().unwrap().is_dir()
			{
				files.push(String::from(name));
			}
		}
		files.sort_unstable();
		assert!(!files.is_empty());

		print!("I found the following source files:");
		for f in &files {
			print!(" {}", style(f).bold().yellow());
		}
		println!();
		println!(
			"{} Which source files would you like to use? (specify multiple files separated by whitespace)",
			style("?").bold().yellow()
		);
		files = ask("files")
			.split_ascii_whitespace()
			.map(String::from)
			.collect();
		assert!(!files.is_empty());

		let project = Project {
			lecture: ProjectLecture {
				course,
				label: preset.label,
				docent: preset.docent,
				date,
				lang: preset.lang
			},
			source: ProjectSource {
				files,
				stereo: args.stereo,
				start: None,
				end: None,
				fast: Vec::new(),
				questions: Vec::new(),
				metadata: None
			},
			progress: Default::default()
		};
		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();
		project
	};

	let renderer = Renderer::new(&directory, &project).unwrap();
	let recording = renderer.recording_mkv();

	// preprocess the video
	if !project.progress.preprocessed {
		renderer.preprocess(&mut project).unwrap();
		project.progress.preprocessed = true;

		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();
	}

	println!();
	println!(
		" {} Preprocessed video: {}",
		style("==>").bold().cyan(),
		style(recording).bold().yellow()
	);

	// ask the user about start and end times
	if !project.progress.asked_start_end {
		println!(
			"{} What is the first/last second you want included?",
			style("?").bold().yellow()
		);
		project.source.start = Some(ask_time("first"));
		project.source.end = Some(ask_time("last "));
		project.progress.asked_start_end = true;

		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();
	}

	// ask the user about fast forward times
	if !project.progress.asked_fast {
		println!(
			"{} Which sections of the video do you want fast-forwarded? (0 to finish)",
			style("?").bold().yellow()
		);
		loop {
			let start = ask_time("from");
			if start.seconds == 0 && start.micros == 0 {
				break;
			}
			let end = ask_time("to  ");
			project.source.fast.push((start, end));
		}
		project.progress.asked_fast = true;

		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();
	}

	// ask the user about questions from the audience that should be subtitled
	if !project.progress.asked_questions {
		println!(
			"{} In which sections of the video were questions asked you want subtitles for? (0 to finish)",
			style("?").bold().yellow()
		);
		loop {
			let start = ask_time("from");
			if start.seconds == 0 && start.micros == 0 {
				break;
			}
			let end = ask_time("to  ");
			let text = ask("text");
			project.source.questions.push((start, end, text));
		}
		project.progress.asked_questions = true;

		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();
	}

	// render the assets
	if !project.progress.rendered_assets {
		renderer.render_assets(&project).unwrap();
		project.progress.rendered_assets = true;

		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();
	}

	// render the video
	let mut videos = Vec::new();
	videos.push(if project.progress.rendered {
		renderer.video_file_output()
	} else {
		let video = renderer.render(&mut project).unwrap();
		project.progress.rendered = true;

		fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes()).unwrap();

		video
	});

	// rescale the video
	if let Some(lowest_res) = args.transcode.or(preset.transcode) {
		for res in Resolution::STANDARD_RESOLUTIONS.into_iter().rev() {
			if res > project.source.metadata.as_ref().unwrap().source_res
				|| res > args.transcode_start.unwrap_or(preset.transcode_start)
				|| res < lowest_res
			{
				continue;
			}
			if !project.progress.transcoded.contains(&res) {
				videos.push(renderer.rescale(&project.lecture, res).unwrap());
				project.progress.transcoded.insert(res);

				fs::write(&project_path, toml::to_string(&project).unwrap().as_bytes())
					.unwrap();
			}
		}
	}

	println!();
	println!(
		" {} {}",
		style("==>").bold().cyan(),
		style("DONE :)").bold()
	);
	println!("     Videos:");
	for v in &videos {
		println!(
			"     {} {}",
			style("->").bold().cyan(),
			style(v).bold().yellow()
		);
	}
}
